import AutoRPE.UtilsRPE.RegexPattern as RegexPattern
import AutoRPE.UtilsRPE.SourceManager as SourceManager
import AutoRPE.UtilsRPE.VariablePrecision as VariablePrecision
import AutoRPE.UtilsRPE.Error as Error
import re


def randStr(N):
    """Generates a random string with specified length"""
    import string
    import random
    chars = string.ascii_uppercase + string.digits
    return ''.join(random.choice(chars) for _ in range(N))


def mask_literal(string, dict_of_substitutions=None):
    """ Search for open and close of different apex and mask what's in between"""
    if dict_of_substitutions is None:
        dict_of_substitutions = {}
    # Before checking for apex, check for apex with escapes
    escaped_string = re.finditer(r'\bb\'.*\'', string)
    for e_s in escaped_string:
        dict_key = randStr(len(e_s.group()))
        dict_of_substitutions[dict_key] = string[e_s.start():e_s.start() + len(dict_key)]
        string = string[:e_s.start()] + dict_key + string[e_s.start() + len(dict_key):]

    for i in [r"\"", r"\\'", r"\'"]:
        apex = re.finditer(i, string)
        apex = [a.start() for a in apex]
        for i_open, i_close in zip(apex[0::2], apex[1::2]):
            dict_key = randStr((i_close - i_open + 1))
            dict_of_substitutions[dict_key] = string[i_open:i_close + 1]
            string = string[:i_open] + dict_key + string[i_close + 1:]
    return string


def mask_keyword(line, key, dictionary_of_substitutions):
    """Check if a variables is called following a keyword with the same name"""
    import re
    key = RegexPattern.escape_string(key)
    match = re.finditer(r'(%s)\s*=\s*%s' % (key, key), line, re.I)
    match = [[m.start(1), m.end(1)] for m in match]
    for start, end in match:
        dict_key = randStr((end - start + 1))
        dictionary_of_substitutions[dict_key] = line[start:end]
        line = line[:start] + dict_key + line[end:]
    return line


def remove_literals(string, dict_of_literal_sub=None):
    if dict_of_literal_sub is None:
        dict_of_literal_sub = {}
    string = mask_literal(string, dict_of_literal_sub)
    # Substitute literals with space, maintaining length
    for key, val in zip(dict_of_literal_sub.keys(), dict_of_literal_sub.values()):
        string = re.sub(RegexPattern.escape_string(key), "\'" + " "*(len(val)-2) + "\'", string)
    return string


def mask_string_for_check_operation(string, dictionary_of_replacements=None):
    import AutoRPE.UtilsRPE.BasicFunctions as BasicFunctions
    # Remove exponent in the form of d**N: it will mess up with normal operators. It will not be substituted back
    string = remove_pow(string)
    string = BasicFunctions.remove_unary_operator(string)
    # Dict to store all entities that will need a mask
    if dictionary_of_replacements is None:
        dictionary_of_replacements = {}
    # Mask the exponential form of numbers/variables (may contain operations in the exponent)
    string = mask_exponential(string, dictionary_of_replacements)

    # Mask all arrays and structures (may contain operation in indexing)
    string = mask_arrays_and_structures(string, dictionary_of_replacements=dictionary_of_replacements)

    # Mask all functions call (may contain operation in arguments)
    string = mask_functions(string, dictionary_of_replacements)

    # Now is safe to remove brackets
    string = string.replace("(", "")
    string = string.replace(")", "")
    return string


def mask_arrays_and_structures(string, only_array=False, only_structures=False, dictionary_of_replacements=None):
    """ Replace all array and structures-like patter, creating a dictionary with the substitutions"""
    import AutoRPE.UtilsRPE.BasicFunctions as BasicFunctions
    import AutoRPE.UtilsRPE.NatureDeterminer as NatureDeterminer
    # Check if the entire string is an array
    if NatureDeterminer.is_array(string) and not only_structures:
        dict_key = randStr(len(string))
        dictionary_of_replacements[dict_key] = string
        return dict_key
    # Start masking
    start = True
    array_replacement = {}
    if dictionary_of_replacements is None:
        # In case we do not want to store the replacement: create a tmp dict
        dictionary_of_replacements = {}
    # Checks for all array pattern, and substitute with randomly generated string of the same length
    while start:
        array_match = re.finditer(RegexPattern.array, string)
        start = [a.start() for a in array_match]
        for match_start in start[::-1]:
            array = BasicFunctions.close_brackets(string[match_start:])
            # Generate a placeholder with specified length
            dict_key = randStr(len(array))
            if not only_structures:
                dictionary_of_replacements[dict_key] = array
            array_replacement[dict_key] = array
            string = string[:match_start] + dict_key + string[match_start + len(array):]

        # Substitute structures with string
        if not only_array:
            struct_match = re.finditer(RegexPattern.structure, string)
            start_end = [[a.start(), a.end()] for a in struct_match]
            for match_start, match_end in start_end[::-1]:
                struct = string[match_start:match_end]
                dict_key = randStr(len(struct))
                dictionary_of_replacements[dict_key] = struct
                string = string[:match_start] + dict_key + string[match_start + len(struct):]
                # Cycle once more to check for arrays/functions after removing structures
                start = True
    if only_structures:
        string = unmask_from_dict(string, array_replacement)
    return string


def remove_pow(string):
    import AutoRPE.UtilsRPE.BasicFunctions as BasicFunctions
    exponentials_var = re.finditer(RegexPattern.number_exponentiation, string)
    # When the var is elevate to simple power, remove power
    start_end = [[a.start(), a.end()] for a in exponentials_var]
    for match_start, match_end in start_end[::-1]:
        string = string[:match_start] + string[match_end:]
    # Exp elevated to some operation in brackets, remove all in brackets
    exponentials_var = re.finditer(r"\*\*\s*\(", string)
    start = [a.start() for a in exponentials_var]
    for match_start in start[::-1]:
        match_end = match_start + len(BasicFunctions.close_brackets(string[match_start:]))
        string = string[:match_start] + string[match_end:]
    return string


def mask_functions(string, dictionary_of_replacements=None):
    """ Once array and structures are masked, this will mask only function"""
    import AutoRPE.UtilsRPE.BasicFunctions as BasicFunctions

    if dictionary_of_replacements is None:
        dictionary_of_replacements = {}

    functions = re.finditer(RegexPattern.call_to_function, string)
    start = [a.start(1) for a in functions]
    for match_start in start[::-1]:
        function_call = BasicFunctions.close_brackets(string[match_start:])
        # Generate a placeholder with specified length
        dict_key = randStr(len(function_call))
        dictionary_of_replacements[dict_key] = function_call
        string = string[:match_start] + dict_key + string[match_start + len(function_call):]
    return string


def mask_exponential(string, dictionary_of_replacements=None):
    """ Replace all exponential creating a dictionary with the substitutions"""
    # If no dictionary is passed, create one
    if dictionary_of_replacements is None:
        dictionary_of_replacements = {}

    exponentials_DPE = re.finditer(RegexPattern.number_exp_DPE_notation, string, re.I)
    # Substitute match with string, group 1, which correspond to the actual exponential
    start_end = [[a.start(2), a.end(2)] for a in exponentials_DPE]
    for match_start, match_end in start_end[::-1]:
        exp_DPE = string[match_start:match_end]
        dict_key = randStr(len(exp_DPE))
        dictionary_of_replacements[dict_key] = exp_DPE
        string = string[:match_start] + dict_key + string[match_start + len(exp_DPE):]
    return string


def mask_intrinsics(string):
    import AutoRPE.UtilsRPE.BasicFunctions as BasicFunctions
    import AutoRPE.UtilsRPE.CallManager as CallManager
    import AutoRPE.UtilsRPE.NatureDeterminer as NatureDeterminer
    intrinsic = NatureDeterminer.has_intrinsics(string)

    while intrinsic:
        # Get the intrinsic call
        intrinsic_call = BasicFunctions.close_brackets(
            string[re.search(RegexPattern.name_occurrence % intrinsic, string, re.I).start():])
        called_arg = CallManager.find_call_arguments(intrinsic_call)
        arguments_to_remove = []
        if intrinsic.lower() in ['real', 'reshape']:
            # Remove loop index, or cast precision if present
            called_arg = called_arg[:1]
        for arg in called_arg:
            if arg.split("=")[0] in SourceManager.list_of_key:
                arguments_to_remove.append(arg)
        # Removed key specification
        for arg in arguments_to_remove:
            called_arg.remove(arg)
        # Rejoin pure arguments
        if len(called_arg) > 1:
            string = string.replace(intrinsic_call, "(" + ",".join(called_arg) + ")", re.I)
        else:
            string = string.replace(intrinsic_call, called_arg[0], re.I)
            string = BasicFunctions.remove_outer_brackets(string)
        # Check for further intrinsics
        intrinsic = NatureDeterminer.has_intrinsics(string)
    # Check if is a cast with precision and the precision was not specified via key
    called_arg = CallManager.find_call_arguments("(" + string + ")")
    if len(called_arg) == 2:
        if called_arg[1] in VariablePrecision.lookup_table:
            return [called_arg[0]]
        # It is the dimension ?
        if NatureDeterminer.is_a_number(called_arg[1]):
            return [called_arg[0]]
    if len(called_arg) > 1:
        raise Error.VariableNotFound("The string %s is not a variable" % string)
    return called_arg


def unmask_from_dict(string, dictionary_of_replacements):
    if not (isinstance(string, list)):
        arguments = [string]
    else:
        arguments = string
    replacement = True
    while replacement:
        replacement = []
        for k in dictionary_of_replacements.keys():
            for a in arguments:
                if re.search(RegexPattern.name_occurrence % k, a):
                    replacement.append(a)
        for key, val in dictionary_of_replacements.items():
            for arg_index, arg in enumerate(arguments):
                if re.search(RegexPattern.name_occurrence % key.strip(), arg):
                    arguments[arg_index] = re.sub(RegexPattern.escape_string(key, strict=False), val, arguments[arg_index])
    if not (isinstance(string, list)):
        return arguments[0]
    else:
        return arguments
